/*******************************************************************************/
/*                                                                             */
/* SPORT SSL for the Blackfin products and silicon revisions                   */
/*                                                                             */
/*     ADSP-BF534-0.1                                                          */
/*     ADSP-BF536-0.1                                                          */
/*     ADSP-BF537-0.1                                                          */
/*                                                                             */
/*     for documentation see manual                                            */
/*                                                                             */
/*        ADSP-BF537 Blackfin Hardware Reference - Section Booting            */
/*                                                                             */
/* (c) Analog Devices Inc. (2006)                                              */
/*                                                                             */
/*******************************************************************************/

#include <defBF537.h>
#define           		SPORT_Descriptors1	0xFF907FC4		
#define           		SPORT_Descriptors2	0xFF907FD8
#define 				SPI_Flash			0xFF907FEC	
#define 				DMA_Destination		0xFF907FF0
#define                 DUMMY_Zeros         0xFF907FFC
#define 				SYSMMR_BASE         0xFFC00000
#define 				COREMMR_BASE        0xFFE00000
#define 				L1_Scratch_END      0XFFB00FFC		//L1 Scratch SRAM (end)
#define                 BEG_APPLICATION_H   0XFFA0
#define                 BEG_APPLICATION_L   0X0000

// Change SSL start address to desired address
//#define                 JMP_LOADER          0x03FFFC60	
//.section jmp_ldr;				// For debugging purposes
//  P0.L = LO(JMP_LOADER);
//  P0.H = HI(JMP_LOADER);
//  JUMP (P0);	

.section program;
.global start;

start:
	CALL Global_Setup;			// Initializes data and pointer core registers	
	CALL DISABLE_SPI;			// Disables SPI in order to use SPORT interface for booting
	CALL INIT_SPORT;			// Initializes SPORT	

GRAB_HEADER:
	CALL SPI_ADDRESS;			// Address passed in R3. Initial value is the beginning SPI address of SSL
	CALL DMA_HEADER;			// Fetch header to intermediate memory
								// R1 = Target Address 
								// R2 = Byte Count
								// R5 = Flags
	R3 += 0xA;					// Increment external address by header

	
	/*******************************************************************************/
	/*                                                                             */
	/*  Test ZEROFILL Flag                                                   .     */
	/*                                                                             */
	/*  If set patch Source Address and start Flash Boot Memory DMA                */
	/*                                                                             */
	/*******************************************************************************/
	
	CHECK_ZERO_FILL_BIT:
		CC = BITTST(R5,0);		
		IF !CC JUMP CHECK_IGNORE_BIT;		// Zero Fill?
	
		CALL ZEROFILL_MDMA;
		JUMP CHECK_LAST_SECTION;			// Check if this is the last block

	
	/*******************************************************************************/
	/*                                                                             */
	/*  Test IGNORE Flag                                                     .     */
	/*                                                                             */
	/*  If set incremenent Source Address Pointer R3 and don't perform any DMA.    */
	/*  Enable INIT flag and FINAL flag to be active at the same time.             */
	/*                                                                             */
	/*******************************************************************************/
	
	CHECK_IGNORE_BIT:
		CC = BITTST(R5,4);					// Ignore ?
		IF CC JUMP DONT_DMA;
	
	
	/*******************************************************************************/
	/*                                                                             */
	/*  DMA payload data in                                                  .     */
	/*                                                                             */
	/*******************************************************************************/	
	
	DO_DMA:
		CALL SPI_ADDRESS;					// Before setting up SPORT DMA, we need 
											// to address the SPI memory	
		CALL SPORT_DMA;						// Perform DMA transfer
			
	DONT_DMA:	
		R3 = R3 + R2;						// increment external address by count 
											// value	
	
	/*******************************************************************************/
	/*                                                                             */
	/*  Test INIT Flag                                                       .     */
	/*                                                                             */
	/*  If set DMA data in and issue a CALL instruction to the target address.     */
	/*  INIT flag and FINAL flag may now coexist in the same block                 */ 
	/*                                                                             */
	/*******************************************************************************/
	
	CHECK_INIT_BIT:
		CC = BITTST(R5,3);
		IF !CC JUMP CHECK_LAST_SECTION;		// Init ?

		P2 = [P3];							// Block's Destination Address
		CALL (P2);
					

	/*******************************************************************************/
	/*                                                                             */
	/*  Test FINAL Flag                                                      .     */
	/*                                                                             */
	/*  Either fetch next block header or exit and JUMP to reset vector            */
	/*                                                                             */
	/*******************************************************************************/										

											
	CHECK_LAST_SECTION:
		CC = BITTST(R5,15);					// Last Section?
		IF !CC JUMP GRAB_HEADER;			// Otherwise parse next block
		JUMP BOOT_END;
	
	CALL BOOT_END;
	
start.END:


/*******************************************************************************/
/*                                                                             */
/*  Global_Setup Routine                                                       */
/*                                                                             */
/*  Initializes most registers as required by various booting procedures.      */
/*  A special entry avois initialization of stack pointers for run-time calls. */
/*                                                                             */
/*******************************************************************************/

Global_Setup:
	
	SP.H = HI(L1_Scratch_END);	// Scratch memory used for stack purposes
	SP.L = LO(L1_Scratch_END);
	FP = SP;
	
	R4 = 0x0000(z);		
	P0 = R4;					// load P0.L
	P1 = R4;					// load P1.L
	M0 = R4;					// value used for zerofill purposes
	M1 = R4;					// MDMA Word Width Indicator to 8-bit 				
	
	P3.H = HI(DMA_Destination);	// P3 Points to the DMA Header Destination.
	P3.L = LO(DMA_Destination); // 0xFF807FF

	P0.H = HI(COREMMR_BASE);	// P0 Points to CORE MMR Space
	P1.H = HI(SYSMMR_BASE);		// P1 Points to SYSTEM MMR Space
	
	P5 = P3;
	P5+= 0x0C;					// 0xFF907FF2
	M2 = P5;					// location for 32-bit zero and trash data
    
	/*Dummy zeros*/
	P5.H = HI(DUMMY_Zeros);
    P5.L = LO(DUMMY_Zeros);    	
	W[P5] = R4.L;						
	
	
/*******************************************************************************/
/*                                                                             */
/*  Global DMA Settings                                                        */
/*     MDMA1 for parallel flash boot and zero fill                             */
/*     DMA7 for SPI boot                                                       */
/*     DMA8 for UART boot                                                      */
/*                                                                             */
/******************************************************************************/																						
	R4 = 0x1;
	W[P1+LO(MDMA_D1_X_MODIFY)] = R4;	// MDMA Destination Modify = 1							
	W[P1+LO(MDMA_S1_X_MODIFY)] = R4; 	


/*******************************************************************************/

	I1 = PF10(Z);				// Set SSEL to PF10 by default

/*******************************************************************************/
   
	R3 = 0x048A(z);				// Init SPI address for application
	R6 = 0x1;                   // Assumes 24-bit SPI memory		
	RTS;
	
Global_Setup.END:

	
/*******************************************************************************/
/*                                                                             */
/*  Disable SPI SSEL output drivers. If there was any other output that		   */
/*  has been enabled by an init code it is not disabled here.                  */
/*                                                                             */
/*******************************************************************************/

DISABLE_SPI:

	[--SP] = RETS;					// Save return address onto the stack
	
	R1 = W[P1+LO(PORTFIO_DIR)] (Z);	// Disable SPI SSEL output driver
	R0 = I1;
	R0 = ~R0;
	R1 = R0 & R1;	
	W[P1+LO(PORTFIO_DIR)] = R1;		
	
	R1 = W[P1+LO(SPI_CTL)] (Z);		// Disables the SPI PORT
	BITCLR(R1,14);
	W[P1+LO(SPI_CTL)] = R1;		

	RETS = [SP++];					// Save return address onto the stack
	
	RTS;

DISABLE_SPI.END:


/*******************************************************************************/
/*                                                                             */
/*  Initialize the SPORT interface										       */
/*                                                                             */
/*******************************************************************************/

INIT_SPORT:
	
	[--SP] = RETS;							// Save return address onto the stack
	[--SP] = R3;

	R3 = 0x0080(Z);							// Only enable SPORT1 RX to wake up 
	[P1 + LO(SIC_IWR)]   = R3;				// the core from idle state
  
	R3.L = (0x3F)<<10;							
	W[P1 + LO(PORTG_FER)] = R3;			

	R3.L = PGTE | PGRE;							
	W[P1 + LO(PORT_MUX)]  = R3;				

	R3.L = 0x0;							
	W[P1 + LO(SPORT1_TCLKDIV)] = R3;			// SPORT CLK = SCLK/2 (35MHz)		
			
	R3.L = 0x7;							
	W[P1 + LO(SPORT1_TFSDIV)]  = R3;			// SPORT TFS every 8 SPORT CLK cycles 	
	
	R3.L = SLEN(0x7);							
	W[P1 + LO(SPORT1_TCR2)]    = R3;			// TX word length 8 bit	
	
	R3.L = TCKFE | LATFS | LTFS | TFSR | ITFS | ITCLK;  
	W[P1 + LO(SPORT1_TCR1)]    = R3;			
	
	R3.L = 0x1;							
	W[P1 + LO(SPORT1_RCLKDIV)] = R3;			// SPORT CLK = SCLK/2 (35MHz)		
			
	R3.L = 0x7;							
	W[P1 + LO(SPORT1_RFSDIV)]  = R3;			// SPORT RFS every 8 SPORT CLK cycles 	
	
	R3.L = SLEN(0x7);							
	W[P1 + LO(SPORT1_RCR2)]    = R3;			// RX word length 8 bit	
	
	R3.L = RCKFE | LARFS | LRFS | RFSR;
	W[P1 + LO(SPORT1_RCR1)]    = R3;				
	
	R3   = [SP++];
	RETS = [SP++]; 
	
	RTS; 
	
INIT_SPORT.END:


/*******************************************************************************/
/*                                                                             */
/*  SPI_ADDRESS Subroutine for SPI Master Boot Mode                            */
/*                                                                             */
/*  Stores READ command and address on SPI_Flash memory space				   */    
/*  Recall:                                                                    */
/*  R6 = 0x3 for 8-bit  SPI Memory                                             */
/*  R6 = 0x2 for 16-bit SPI Memory                                             */
/*  R6 = 0x1 for 24-bit SPI Memory                                             */
/*  R6 = 0x0 for ATMEL DataFlash Memory                                        */
/*                                                                             */
/*******************************************************************************/

SPI_ADDRESS:
	[--SP] = RETS;					// Save return address onto the stack

	P5.H = HI(SPI_Flash);			// P6 Points to the SPI_Flash Commands.
	P5.L = LO(SPI_Flash); 			// 0xFF807FF0
	I0 = R3;						// save logical address in I0
	CC = R6 < 1;
	IF !CC JUMP ADDRESS_8_16_24;
	
	R0 = 0xE8;						

	B[P5++] =  R0;					// Storing read command	

	JUMP ADDRESS_24;	
	
ADDRESS_8_16_24:					// normal SPI devices
	R0 = 0x3;						// control word for 8-/16-/24-bit SPI Memory

	B[P5++] =  R0;					// Storing read command	

	CC = R6 < 2;
	IF !CC JUMP ADDRESS_16;
	
ADDRESS_24:
	R0 = R3 >> 16;					// Storing upper address byte

	B[P5++] =  R0;					
	
ADDRESS_16:
	CC = R6 < 3;
	IF !CC JUMP ADDRESS_8;

	R0 = R3 >> 8;					// Storing middle address byte
	
	B[P5++] =  R0;					

ADDRESS_8:
	R0 = R3;						// Storing lower address byte
	
	B[P5++] =  R0;					

	CC = R6 < 1;
	IF !CC JUMP SPI_ADDRESS.EXIT;
											 
	P2 = 0x4;						// send out 4 don't care bytes	
	LSETUP(ATMEL_Begin, ATMEL_End) LC0=P2;	// for ATMEL SPI Memory
	ATMEL_Begin:	R0 = R3;								
				B[P5++] =  R0;					
	ATMEL_End:		NOP;
	
	R3 = I0;						// restore logical address from I0	


SPI_ADDRESS.EXIT:		
	RETS = [SP++];					// Restore return address from the stack
	RTS;

SPI_ADDRESS.END:


/*******************************************************************************/
/*                                                                             */
/*  DMA_HEADER Subroutine for SPI Master and Slave Boot Mode                   */
/*                                                                             */
/*  It fetches the next 10-byte header in by DMA. L1 memory is used for        */
/*  intermediate storage, but then result is copied to registers:              */
/*                                                                             */
/*  R1 = Target Address (4 bytes)                                              */
/*  R2 = Byte Count (4 bytes)                                                  */
/*  R5 = Flags (2 bytes)                                                       */
/*                                                                             */
/*******************************************************************************/

DMA_HEADER:

	[--SP] = RETS;
	
	R1 = P3;						// DMA Destination = 0xFF807FF0
	R2 = 0xA;						// Block Header size = 0xA									
	
	CALL SPORT_DMA;					// instruction order because of 
									// anomaly #05000209											
											
	R5.L = 0x1001;					// Deposit 1 bit at bit position = 16	
	
	R1 = [P3];						// Target Address 
	R2 = [P3+0x4];					// Byte Count
	R0 = W[P3+0x8](Z);				// Flags
	
	R5 = DEPOSIT(R0, R5);			// After this instruction R5.L = R0.L  
									// and bit0 of R5.H will be retrieved
	RETS = [SP++];
	RTS;
	
DMA_HEADER.END:	



/*******************************************************************************/
/*                                                                             */
/*  SPI_DMA Subroutine for SPI Master, SPI Slave and UART Slave Boot Mode      */
/*                                                                             */
/*  It starts a SPI DMA and wait until it completes                            */
/*  It supports 32-bit byte counts including from 0                            */
/*                                                                             */
/*  R1 = Destination Start Address                                             */
/*       Updated by the routine (R1 = R1 + R2)                                 */
/*                                                                             */
/*  R2 = Byte Count (0..2^32-1)                                                */
/*                                                                             */
/*                                                                             */
/*  R4 = SPI CTL register value                                                */
/*       SPI master: R4=SPE | MSTR | (TIMOD & 2)                               */
/*       SPI slave:  R4=SPE | CPHA | (TIMOD & 2)                               */
/*       UART:       R4=0                                                      */
/*                                                                             */
/*  P4 = DMA Pointer                                                           */
/*       SPI master: P4=DMA7_CONFIG                                            */
/*       SPI slave:  P4=DMA7_CONFIG                                            */
/*       UART:       P4=DMA8_CONFIG                                            */
/*                                                                             */
/*  P5 = HWAIT Toggle Register                                                 */
/*  M3 = HWAIT GPIO Number                                                     */
/*                                                                             */
/*                                                                             */
/*******************************************************************************/


SPORT_DMA:

	[--SP] = (R7:2);	

	P4.H = HI(SPI_Flash);				// P6 Points to the SPI_Flash Commands.
	P4.L = LO(SPI_Flash); 				// 0xFF807FF0
	P5.H = HI(SPORT_Descriptors1);		// P7 Points to the SPORT_Descriptors
	P5.L = LO(SPORT_Descriptors1); 		// 0xFF807FD0

	SSYNC;

	//Initializing descriptors
	R7.L    = 0x4;						// SPI Opcode + Address count
	R3.L    = (FLOW_ARRAY | NDSIZE_5 | WDSIZE_8 | DMAEN);	// Transmitter  
	 									// DMA Enable, 8-bit Memory Read,  
	 									// 1-D DMA, Flow = Array, 	
	R5.L    = 0x1;						// X_MOD
			
	R6 = P4;
	W[P5] = R6.L;							// MOSI1 start_address_l	
	P5+=2;
	W[P5] = R6.H;							// MOSI1 start_address_h	
	P5+=2;
	W[P5++] = R3;							// MOSI1 config
	W[P5++] = R7;							// MOSI1 count
	W[P5++] = R5;							// MOSI1 xmod

	R3.L = (FLOW_STOP | NDSIZE_0 | WDSIZE_8 | DMAEN);	// Transmitter  
	 									// DMA Enable, 8-bit Memory Read,  
	 									// 1-D DMA, Flow = Stop, 

	R5.L    = 0x0;							// X_MOD
	 								        // Sending dummy zeros
    R6.H = HI(DUMMY_Zeros);
    R6.L = LO(DUMMY_Zeros);	
	
	W[P5] = R6.L;							// MOSI1 start_address_l
	P5+=2;
	W[P5] = R6.H;							// MOSI1 start_address_h
	P5+=2;
	W[P5++] = R3;							// MOSI2 config
	W[P5++] = R2;							// MOSI2 count
	W[P5++] = R5;							// MOSI2 xmod

	P5.H = HI(SPORT_Descriptors2);		// P7 Points to the SPORT_Descriptors
	P5.L = LO(SPORT_Descriptors2); 		// 0xFF807FD0

	
	R3.L = (FLOW_ARRAY | NDSIZE_5 | WNR | WDSIZE_8 | DMAEN);// Receiver  
	 									// DMA Enable, 8-bit Memory Write,  
	 									// 1-D DMA, Flow = Array, 

	R5.L    = 0x0;							// X_MOD

	W[P5] = R1.L;							// MISO1 start_address_l
	P5+=2;
	W[P5] = R1.H;							// MISO1 start_address_h
	P5+=2;
	W[P5++] = R3;							// MISO1 config
	W[P5++] = R7;							// MISO1 count
	W[P5++] = R5;							// MISO1 xmod

	R3.L = (FLOW_STOP | NDSIZE_0 | DI_EN | WNR | WDSIZE_8 | DMAEN);//Receiver  
	 									// DMA Enable, 8-bit Memory Write,  
	 									// 1-D DMA, Flow = Stop, 

	R5.L    = 0x1;							// X_MOD

	W[P5] = R1.L;							// MISO1 start_address_l
	P5+=2;
	W[P5] = R1.H;							// MISO1 start_address_h
	P5+=2;	
	W[P5++] = R3;							// MISO2 config
	W[P5++] = R2;							// MISO2 count
	W[P5++] = R5;							// MISO2 xmod			

	//now DMA
	//transmitter
	P4.H = HI(DMA6_CONFIG);				
	P4.L = LO(DMA6_CONFIG); 			
	P5.H = HI(SPORT_Descriptors1);					// P7 Points to the SPI_Descriptors
	P5.L = LO(SPORT_Descriptors1); 					// 0xFF807FD0

	R3.L = 0x0;						
	W[P4+DMA6_CONFIG-DMA6_CONFIG] = R3;				

	R3.L = 0x6000;						
	W[P4+DMA6_PERIPHERAL_MAP-DMA6_CONFIG] = R3;				
	
	[P4+DMA6_CURR_DESC_PTR-DMA6_CONFIG] = P5;				

	R3.L = (FLOW_ARRAY | NDSIZE_5 | WDSIZE_8 | DMAEN);
	W[P4+DMA6_CONFIG-DMA6_CONFIG] = R3;				

	//receiver
	P4.H = HI(DMA5_CONFIG);				
	P4.L = LO(DMA5_CONFIG); 			
	P5.H = HI(SPORT_Descriptors2);					// P7 Points to the SPI_Descriptors
	P5.L = LO(SPORT_Descriptors2); 					// 0xFF807FD0

	R3.L = 0x0;						
	W[P4+DMA5_CONFIG-DMA5_CONFIG] = R3;				

	R3.L = 0x5000;						
	W[P4+DMA5_PERIPHERAL_MAP-DMA5_CONFIG] = R3;				
	
	[P4+DMA5_CURR_DESC_PTR-DMA5_CONFIG] = P5;				

	R3.L = (FLOW_ARRAY | NDSIZE_5 | WDSIZE_8 | DMAEN | WNR);
	W[P4+DMA5_CONFIG-DMA5_CONFIG] = R3;				

	SSYNC;

	/* Enable Sport1 RX and TX */

	P5.H = HI(SPORT1_RCR1);
	P5.L = LO(SPORT1_RCR1);
	R1 = W[P5](Z);
	BITSET(R1,0);
	W[P5] = R1;
	
	P5.H = HI(SPORT1_TCR1);
	P5.L = LO(SPORT1_TCR1);
	R1 = W[P5](Z);
	BITSET(R1,0);
	W[P5] = R1;
	SSYNC;      /* Enable Transmitter (set bit 0) */

	IDLE;			// Wait for DMA to Complete

	R3 = 0x1;
	W[P4+DMA5_IRQ_STATUS-DMA5_CONFIG] = R3;	// Write 1 to clear DMA interrupt
	P4.H = HI(DMA6_CONFIG);				
	P4.L = LO(DMA6_CONFIG); 			
	W[P4+DMA6_IRQ_STATUS-DMA6_CONFIG] = R3;	// Write 1 to clear DMA interrupt

	/* Disable Sport1 RX and TX */

	P5.H = HI(SPORT1_TCR1);
	P5.L = LO(SPORT1_TCR1);
	R1 = W[P5](Z);
	BITCLR(R1,0);
	W[P5] = R1;
	SSYNC;      /* Disable Transmitter (set bit 0) */

	P5.H = HI(SPORT1_RCR1);
	P5.L = LO(SPORT1_RCR1);
	R1 = W[P5](Z);
	BITCLR(R1,0);
	W[P5] = R1;
	SSYNC;      /* Disable Receiver (set bit 0) */

	(R7:2) = [SP++];
	RTS;
	
SPI_DMA.END:


/*******************************************************************************/
/*                                                                             */
/*  ZEROFILL_MDMA Subroutine for all Boot Modes                          .     */
/*                                                                             */
/*  When the ZEROFILL flag is set by any boot mode this function is called.    */
/*  It simply patches the Source channel of the MDMA and calls MEMDMA then.    */
/*                                                                             */
/*  This operation uses MDMA0 always. Note that in FIFO mode the MDMA1 is      */
/*  gated by the DMAR1 and cannot be used for this purpose, therefore.         */
/*                                                                             */
/*  The ZERO variable is no longer an 8-bit constant in ROM. Rather it is a    */
/*  16-bit value in RAM. This way, the ZEROFILL functionality can be abused    */
/*  to full memory with other byte values than zero, when an INIT code patches */
/*  the ZERO RAM location. M0 is expected to hold the value for zerofill DMA.  */
/*                                                                             */
/*  Having ZERO located within L1 memory, reduces speed of zero-filling of L1  */
/*  memories, but slightly improves zero-filling to external memory, which may */
/*  be much bigger in size.                                                    */
/*                                                                             */
/*******************************************************************************/	


ZEROFILL_MDMA:

	[--SP] = (R7:0,P5:4);
	[--SP] = M3;
	[--SP] = RETS;
	
	R0 = 0x0;
	M3 = R0;								// temporary disable HWAIT toggle
	W[P1+LO(MDMA_S0_X_MODIFY)] = R0;		// Source Modify = 0 for Zero Fill
	
	R0 = W[P1+LO(MDMA_D1_X_MODIFY)] (Z);	// copy modify value from MDMA1
	W[P1+LO(MDMA_D0_X_MODIFY)] = R0;
	
	P4 = M2;								// Pointer to ZERO Location
	R0 = M2;								// Source Base Address = ZERO 	
	
	R7 = M0;								// source value used for zerofill
	[P4] = R7;								// defaults to zero if not 
											// patched by init codes

	P4 = P1;
	P4.L = LO(MDMA_S0_CONFIG);				// MDMA routine uses MDMA0
	
	CALL MEMDMA;	
	
	RETS = [SP++]; 
	M3 = [SP++]; 
	(R7:0,P5:4) = [SP++]; 				
	
	RTS;

ZEROFILL_MDMA.END:



/*******************************************************************************/
/*                                                                             */
/*  FDMA/MEMDMA Subroutine for Flash Boot Mode (subroutine to all boot modes)  */
/*                                                                             */
/*  Sets up a MDMA sequence and waits until MDMA is done.                      */
/*  This functions is used to fetch block headers, to boot payload data, as    */
/*  well as to perform zero-initialization. It supports 32-bit byte counts     */
/*  including 0.                                                               */
/*                                                                             */
/*  R0 = Start Source Address                                                  */
/*       returns updated as R0 = R0 + (R2 << M1)                               */
/*                                                                             */
/*  R1 = Start Destination Address                                             */
/*       returns updated as R1 = R1 + (R2 << M1)                               */
/*                                                                             */
/*  R2 = Byte Count                                                            */
/*       is internally divided by 2 in 16-bit DMA mode                         */
/*                                                                             */
/*  M1 = 16-bit DMA Indicator                                                  */
/*       0: 8-bit DMA                                                          */
/*       1: 16-bit DMA                                                         */
/*       2: 32-bit DMA                                                         */
/*                                                                             */
/*  M3 = HWAIT GPIO Number                                                     */
/*                                                                             */
/*  P1 = System MMR base                                                       */
/*       always 0xFFC0.0000                                                    */
/*                                                                             */
/*  P4 = Pointer to either MDMA_S0_CONFIG or MDMA_S1_CONFIG. Determines        */
/*       whether MDMA0 or MDMA1 is used. The FDMA entry overwrites P4          */
/*       to MDMA_S1_CONFIG always.                                             */
/*                                                                             */
/*  P5 = Pointer to GPIO Toggle register are dummy memory address              */
/*                                                                             */
/*                                                                             */
/*  Note that both MDMA X Modify register must be set by the calling routine.  */
/*                                                                             */
/*******************************************************************************/	


FDMA:										// entry for legacy calls

	[--SP] = (R7:2,P5:4);

	P4 = P1;								// whether MDMA0 or MDMA1
	P4.L = LO(MDMA_S1_CONFIG);
	
	JUMP MEMDMA.ENTRY;						// bypass push operation

MEMDMA:										// entry for ZEROFILL calls
	[--SP] = (R7:2,P5:4);

	
MEMDMA.ENTRY:	
	
	R6.H = 0x6000;							// Only enable MemDMA to wake up 
	R6.L = 0x0;								// the core from idle state
	[P1 + LO(SIC_IWR)] = R6;
	
	R4.L = DI_EN | WDSIZE_8 | WNR | DMAEN;	// Destination DMA config
											// 0x0083 = DMA Enable,  Memory Write
											// 8-Bit Transfers, 1-D DMA, 
											// Flow = Stop, Interrupt on 
											// Completion

	R6 = DMAEN;								// Source DMA config
	
/*******************************************************************************/	
/*                                                                             */
/*  If M1=0 then perform  8-bit DMA                                            */
/*  If M1=1 then perform 16-bit DMA                                            */
/*  If M1=2 then perform 32-bit DMA                                            */
/*                                                                             */
/*******************************************************************************/	
	
	R5 = M1;			// M1 = 0 normally, but 1 for true 16-bit flash
	R2 >>= R5;			// half transfer count
	
	R5 <<= 2;			
	R4 = R4 | R5;		// enable 16-bit DMA operation for destination channel
	R7 = R5 | R6;		// enable 16-bit DMA operation for source channel
										
	
/*******************************************************************************/	
/*                                                                             */
/*  To support byte counts greater than 16 bit, multiple DMA sequences might   */
/*  required. The first pass performs the 16-bit reminder, afterward all DMAs  */
/*  have a transfer count of 65536.                                            */
/*                                                                             */
/*  Transfer counts of zero bypass the sequence as no action is required.      */
/*                                                                             */
/*******************************************************************************/	

	R6 <<= 16;							// R6 = 0x0001.0000	
	R5 = R2.L (Z);						// R5.L = R2.L
	CC = R5 == 0;						// R5.H = 0   IF R2.L != 0
	IF CC R5 = R6;						// R5.H = 1   IF R2.L == 0	
										// R6 used below again.
										
MEMDMA.NEXT:	
	CC = R2 == 0;
	IF CC JUMP MEMDMA.EXIT;				// If the COUNT = 0, skip the DMA
		
	SSYNC;
	
	
/*******************************************************************************/
/*                                                                             */
/*  Set Source and Destination settings as calculated above.                   */
/*  Destination DMA issues interrupt when done. 1D-DMA, Flow = STOP            */
/*                                                                             */
/*******************************************************************************/

	[P4+MDMA_S1_START_ADDR-MDMA_S1_CONFIG] = R0;	// Set Source Address
	[P4+MDMA_D1_START_ADDR-MDMA_S1_CONFIG] = R1;	// Set Destination Address

	W[P4+MDMA_S1_X_COUNT-MDMA_S1_CONFIG] = R5;		// Set Source Count
	W[P4+MDMA_D1_X_COUNT-MDMA_S1_CONFIG] = R5;		// Set Destination Count

	W[P4+MDMA_S1_CONFIG-MDMA_S1_CONFIG] = R7;		// Enable Source DMA  						
	W[P4+MDMA_D1_CONFIG-MDMA_S1_CONFIG] = R4;		// Enable Destination DMA  

		
/*******************************************************************************/
/*                                                                             */
/*  Wait for MDMA to complete. IRQ of Destination channels wakes core up.      */
/*  Clear interrupt request afterward.                                         */
/*                                                                             */
/*******************************************************************************/
	
	IDLE;										// Wait for DMA to Complete

	R3 = 1;										// W1C to clear IRQ request
	W[P4+MDMA_D1_IRQ_STATUS-MDMA_S1_CONFIG] = R3;	

	
/*******************************************************************************/
/*                                                                             */
/*  If the total transfer count was greater than 64k the DMA will be started   */
/*  again. Next start addresses and count values need to be calculated for the */
/*  next run. Since both, source and destination DMA may have modify values of */
/*  0, 1 or 2, it is simpler to read the CURR_ADDR registers than calculating  */
/*  the new address by math.                                                   */
/*                                                                             */
/*  If the count was greater than 64k then the first run processed the         */
/*  reminder, if there was any. All further DMA sequences perform full 64k     */
/*  transfer count.                                                            */
/*                                                                             */
/*  Funny instruction order due to anomaly 05000198, never saw it fail however */
/*                                                                             */
/*******************************************************************************/
	
	R2 = R2 - R5;								// Decrement Transfer Count
	
	R0 = [P4+MDMA_S1_CURR_ADDR-MDMA_S1_CONFIG];	// update address pointer

	R5 = R6;									// count = 64k next time
										
	R1 = [P4+MDMA_D1_CURR_ADDR-MDMA_S1_CONFIG];	// update address pointer
		
	JUMP MEMDMA.NEXT;							// next DMA sequence
	
MEMDMA.EXIT:

	//Let SPORT inteface be the only one that can interrupt the core
	R6 = 0x0080(Z);							// Only enable SPORT1 RX to wake up 
	[P1 + LO(SIC_IWR)]   = R6;				// the core from idle state

	(R7:2, P5:4) = [SP++]; 	

	RTS;
	
MEMDMA.END:	
FDMA.END:

/*******************************************************************************/
/*                                                                             */
/*  BOOT_END routine for all boot modes.                                       */
/*                                                                             */
/*  Once Boot Process has finished, some housecleaning is required. It is      */
/*  basically the restoring of reset values to most control registers touched  */
/*  during booting.                                                            */
/*                                                                             */
/*  The second piece is also executed on software resets that don't require    */
/*  rebooting.                                                                 */
/*                                                                             */
/*  Finally, JUMP to reset vector stored in EVT1 to exit. Recall the           */
/*  processor is still running at reset priority.                              */
/*                                                                             */
/*******************************************************************************/

BOOT_END:
	
    R1 = DRQ_MULTI;
	W[P1+LO(HMDMA1_CONTROL)] = R1;		// restore reset value
	
	R1 = 0x00F2;						// disable upper banks and SCLK again
	W[P1+LO(EBIU_AMGCTL)] = R1;

	R1 = 0x0;
	W[P1+LO(SPI_CTL)] = R1;
	W[P1+LO(UART0_GCTL)] = R1;
	W[P1+LO(UART0_IER)] = R1;
	W[P1+LO(TWI_CONTROL)] = R1;	
	W[P1+LO(MDMA_S0_CONFIG)] = R1;
	W[P1+LO(MDMA_D0_CONFIG)] = R1;
	W[P1+LO(MDMA_S1_CONFIG)] = R1;
	W[P1+LO(MDMA_D1_CONFIG)] = R1;
	
	W[P4] = R1; 						// either DMA7_CONFIG or DMA8_CONFIG
	
	W[P1+LO(PORTF_FER)] = R1;
	W[P1+LO(PORT_MUX)] = R1;
		
	R1 = ~R1;
	[P1+LO(SIC_IWR)] = R1;				// reset SIC_IWR to all FF's

	p0.h = hi(EVT1);
    p0.l = lo(EVT1);
    r0.h = BEG_APPLICATION_H;  
    r0.l = BEG_APPLICATION_L;  
    [p0] = r0;
    ssync;
/************************************/
	

/*******************************************************************************/
/*                                                                             */
/*  Jump to reset vector stored in EVT1 register. By default this is           */
/*  0xFFA0.0000, but it can be patched by init codes as well as by user        */
/*  programs in case of software resets.                                       */
/*                                                                             */
/*******************************************************************************/
	
	
SWRESET:	
	
	P0.L = LO(EVT1);
	P0.H = HI(EVT1);
	P1 = [P0]; 	
										// the Event Vector Table
	JUMP(P1);							// JUMP to the RESET vector

BOOT_END.END:

